@@ -1,5 +1,5 @@
 /* PSPPIRE - a graphical user interface for PSPP.
-   Copyright (C) 2008, 2009, 2010  Free Software Foundation
+   Copyright (C) 2008, 2009, 2010, 2011  Free Software Foundation
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -24,9 +24,7 @@
 #include <gtk/gtk.h>
 #include "psppire-window.h"
 #include "psppire.h"
-
-extern int viewer_length;
-extern int viewer_width ;
+#include "libpspp/string-map.h"
 
 
 G_BEGIN_DECLS
@@ -55,6 +53,7 @@ struct _PsppireOutputWindow
   int max_width;
   int y;
 
+  struct string_map render_opts;
   GtkTreeView *overview;
   GtkTreeIter cur_command;
   bool in_command;
@@ -67,6 +66,9 @@ struct _PsppireOutputWindow
   struct xr_driver *print_xrd;
   int print_item;
   int print_n_pages;
+  gboolean paginated;
+
+  gboolean dispose_has_run;
 };
 
 struct _PsppireOutputWindowClass
